const {assign, create, freeze, defineProperties} = Object;

/**
 * Create initialized parse state.
 * @param {Interface} cli
 * @param {string[]} args
 * @returns {State}
 */
function State(cli, args) {
    const index = 0;
    const prev = null;

    args = freeze(args.map(String));

    return freeze(assign(create(State.prototype), {cli, index, prev, args}));
}

defineProperties(State.prototype, {
    /**
     * Current argument.
     * @name State#curr
     * @type {string}
     * @readonly
     */
    curr: {
        configurable: true,
        enumerable: true,
        get: function() {
            return this.args[this.index];
        }
    },

    /**
     * Lookahead.
     * @name State#ahead
     * @type {string[]}
     * @readonly
     */
    ahead: {
        configurable: true,
        enumerable: true,
        get: function() {
            return freeze(this.args.slice(this.index));
        }
    },

    /**
     * Lookbehind.
     * @name State#behind
     * @type {string[]}
     * @readonly
     */
    behind: {
        configurable: true,
        enumerable: true,
        get: function() {
            return freeze(this.args.slice(0, this.index).reverse());
        }
    }
});

/**
 * Update state.
 * @param {object} delta
 * @returns {State}
 */
State.prototype.update = function(delta) {
    var state = create(State.prototype);
    return freeze(assign(state, this, {prev: this}, delta));
};

/**
 * Shift argument.
 * @returns {State}
 */
State.prototype.shift = function() {
    return this.index < this.args.length
        ? this.update({index: this.index + 1})
        : null;
};

/**
 * Expand arguments.
 * @param {number} [num=1]
 * @param {string[]} expansion
 * @returns {State}
 */
State.prototype.expand = function(num, expansion) {
    if (arguments.length === 1) expansion = num, num = 1;

    var args,
        expanded = null;

    if (this.index + num <= this.args.length) {
        args = this.args.slice(0, this.index)
            .concat(expansion)
            .concat(this.args.slice(this.index + num));

        expanded = this.update({args: freeze(args)});
    }

    return expanded;
};

/**
 * True if parsing has finished.
 * @returns {boolean}
 */
State.prototype.parsed = function() {
    return this.index === this.args.length;
};

/**
 * Return copy of remaining arguments.
 * @returns {string[]}
 */
State.prototype.rest = function() {
    return this.args.slice(this.index);
};

module.exports = State;
