const {defineProperties, isFrozen} = Object;
const expect = require("expect.js");
const sinon = require("sinon");
const CLI = require("..");

describe("qwcli.State", () => {
	describe("(Interface, string[])", () => {
		it("should return an initialized State", () => {
			const cli = {};
			const args = ["foo", "bar"];
			const state = CLI.State(cli, args);

			expect(state).to.be.a(CLI.State);
			expect(state.cli).to.be(cli);
			expect(state.args).to.be.an("array");
			expect(state.args.length).to.be(2);
			expect(state.args[0]).to.be("foo");
			expect(state.args[1]).to.be("bar");
			expect(state.index).to.be(0);
			expect(state.prev).to.be(null);
			expect(isFrozen(state)).to.be(true);
		});
	});

	describe("#update(object)", () => {
		var state, updated, delta;

		beforeEach(() => {
			state = CLI.State({}, ["foo"]);
			delta = {index: 1};
			updated = state.update(delta);
		});

		it("should return a new State", () => {
			expect(updated).to.be.a(CLI.State);
			expect(updated).to.not.be(state);
		});

		it("should update state with delta object", () => {
			expect(updated.index).to.be(1);
		});

		it("should set prev to previous state", () => {
			expect(updated.prev).to.be(state);
		});

		it("should allow delta to override prev", () => {
			expect(state.update({prev: updated}).prev).to.be(updated);
		});
	});

	describe("#shift()", () => {
		var state, shifted;

		beforeEach(() => {
			state = CLI.State({}, ["foo"]);
			shifted = state.shift();
		});

		it("should return state with index advanced by one", () => {
			expect(shifted).to.be.a(CLI.State);
			expect(shifted.index).to.be(1);
		});

		it("should return null if already at end of args", () => {
			expect(shifted.shift()).to.be(null);
		});
	});

	describe("#expand(string[]", () => {
		const expansion = ["baz", "bang"];
		var state, expanded;

		beforeEach(() => {
			state = CLI.State({}, ["foo", "bar"]);
			expanded = state.expand(expansion);
		});

		it("should return state with argument expanded", () => {
			expect(expanded).to.be.a(CLI.State);
			expect(expanded.args.length).to.be(3);
			expect(expanded.args[0]).to.be(expansion[0]);
			expect(expanded.args[1]).to.be(expansion[1]);
			expect(expanded.args[2]).to.be(state.args[1]);
		});
	});

	describe("#expand(number, string[])", () => {
		const expansion = ["baz", "bang"];
		var state, expanded;

		beforeEach(() => {
			state = CLI.State({}, ["foo", "bar", "biff"]);
			expanded = state.expand(2, expansion);
		});

		it("should return state with multiple args expanded", () => {
			expect(expanded).to.be.a(CLI.State);
			expect(expanded.args.length).to.be(3);
			expect(expanded.args[0]).to.be(expansion[0]);
			expect(expanded.args[1]).to.be(expansion[1]);
			expect(expanded.args[2]).to.be(state.args[2]);
		});
	});

	describe("#parsed()", () => {
		var state;

		beforeEach(() => {
			state = CLI.State({}, ["foo"]);
		});

		it("should return true when completely parsed", () => {
			expect(state.parsed()).to.be(false);
			expect(state.shift().parsed()).to.be(true);
		});
	});

	describe("#rest()", () => {
		var state;

		beforeEach(() => {
			state = CLI.State({}, ["foo"]);
		});

		it("should return copy of remaining arguments", () => {
			expect(state.rest()).to.be.an("array");
			expect(state.rest()).to.not.be(state.args);
			expect(state.rest()).to.not.be(state.rest());
			expect(state.rest().length).to.be(1);
			expect(state.rest()[0]).to.be("foo");
			expect(state.shift().rest().length).to.be(0);
		});
	});

	describe("#curr", () => {
		var state;

		beforeEach(() => {
			state = CLI.State({}, ["foo", "bar"]);
		});

		it("should be set to the current argument", () => {
			expect(state.curr).to.be("foo");
			expect(state.shift().curr).to.be("bar");
			expect(state.shift().shift().curr).to.be(undefined);
		});
	});

	describe("#ahead", () => {
		var state;

		beforeEach(() => {
			state = CLI.State({}, ["foo", "bar"]);
		});

		it("should be unique each time it is read", () => {
			// TODO: fix this; this is dumb
			expect(state.ahead).to.not.be(state.ahead);
		});

		it("should be a frozen array", () => {
			expect(state.ahead).to.be.an("array");
			expect(isFrozen(state.ahead)).to.be(true);
		});

		it("should contain current and upcoming arguments", () => {
			expect(state.ahead[0]).to.be("foo");
			expect(state.ahead[1]).to.be("bar");
			expect(state.shift().ahead[0]).to.be("bar");
		});
	});

	describe("#behind", () => {
		var state;

		beforeEach(() => {
			state = CLI.State({}, ["foo", "bar"]);
		});

		it("should be unique each time it is read", () => {
			// TODO: fix this; this is dumb
			expect(state.behind).to.not.be(state.behind);
		});

		it("should be a frozen array", () => {
			expect(state.behind).to.be.an("array");
			expect(isFrozen(state.behind)).to.be(true);
		});

		it("should contain parsed arguments in reverse order", () => {
			expect(state.behind.length).to.be(0);
			expect(state.shift().behind.length).to.be(1);
			expect(state.shift().shift().behind.length).to.be(2);			
			expect(state.shift().shift().behind[0]).to.be("bar");
			expect(state.shift().shift().behind[1]).to.be("foo");
		});
	});
});

describe("qwcli.Interface", () => {
	var cli;

	beforeEach(() => {
		cli = new CLI.Interface();
	});

	describe("#parser()", () => {
		it("should return a parse function", () => {
			expect(cli.parser()).to.be.a("function");
		});
	});

	describe("#parser() => function(array)", () => {
		it("should return unparsed arguments", () => {
			const parse = cli.parser();
			const args = ["foo", "bar"];
			const result = parse(args);

			expect(result).to.not.be(args);
			expect(result).to.be.an("array");
			expect(result.length).to.be(2);
			expect(result[0]).to.be("foo");
			expect(result[1]).to.be("bar");
		});
	});

	describe("#bind(string, function)", () => {
		it("should return binding symbol", () => {
			const spy = sinon.spy();
			expect(typeof cli.bind("foo", spy)).to.be("symbol");
		});

		it("should bind option to handler", () => {
			const spy = sinon.spy();

			cli.bind("foo", spy);
			cli.parser()(["foo"]);

			expect(spy.called).to.be(true);
		});

		it("should pass positional arguments to handler", () => {
			const spy = defineProperties(sinon.spy(), {length: {value: 2}});

			cli.bind("foo", spy);
			const argv = cli.parser()(["foo", "bar", "baz", "bang"]);

			expect(spy.called).to.be(true);
			expect(spy.calledWith("bar", "baz")).to.be(true);
			expect(argv.length).to.be(1);
			expect(argv[0]).to.be("bang");
		});

		it("should error on attempt to set option twice", () => {
			const spy = sinon.spy();

			cli.bind("foo", spy);
			expect(() => cli.bind("foo", spy)).to.throwError();
		});
	});

	describe("#bind(string[], function)", () => {
		it("should return binding symbol", () => {
			const spy = sinon.spy();
			expect(typeof cli.bind(["foo"], spy)).to.be("symbol");
		});

		it("should bind option aliases to handler", () => {
			const spy = sinon.spy();

			cli.bind(["foo", "bar"], spy);
			cli.parser()(["bar"]);

			expect(spy.called).to.be(true);
		});

		it("should pass positional arguments to handler", () => {
			const spy = defineProperties(sinon.spy(), {length: {value: 2}});

			cli.bind(["foo", "bar"], spy);
			const argv = cli.parser()(["bar", "foo", "baz", "bang"]);

			expect(spy.called).to.be(true);
			expect(spy.calledWith("foo", "baz")).to.be(true);
			expect(argv.length).to.be(1);
			expect(argv[0]).to.be("bang");
		});

		it("should error on attempt to set option twice", () => {
			const spy = sinon.spy();

			cli.bind(["foo", "bar"], spy);
			expect(() => cli.bind("foo", spy)).to.throwError();
			expect(() => cli.bind("bar", spy)).to.throwError();
		});
	});

	describe("#bind(qwcli.head, function)", () => {
		it("should return binding symbol", () => {
			const spy = sinon.spy();
			expect(typeof cli.bind(CLI.head, spy)).to.be("symbol");
		});

		it("should bind positional arguments to handler", () => {
			const spy = defineProperties(sinon.spy(), {length: {value: 2}});

			cli.bind(CLI.head, spy);
			const argv = cli.parser()(["baz", "bang", "biff"]);

			expect(spy.called).to.be(true);
			expect(spy.calledWith("baz", "bang")).to.be(true);
			expect(argv.length).to.be(1);
			expect(argv[0]).to.be("biff");
		});

		it("should bind positional arguments before options", () => {
			const headspy = defineProperties(sinon.spy(), {length: {value: 1}});
			const foospy = sinon.spy();
			const barspy = sinon.spy();

			cli.bind(CLI.head, headspy);
			cli.bind("foo", foospy);
			cli.bind("bar", barspy);
			const argv = cli.parser()(["foo", "bar", "baz"]);

			expect(headspy.called).to.be(true);
			expect(headspy.calledWith("foo")).to.be(true);
			expect(foospy.called).to.be(false);
			expect(barspy.called).to.be(true);
			expect(argv.length).to.be(1);
			expect(argv[0]).to.be("baz");
		});

		it("should error on attempt to set head twice", () => {
			const spy = defineProperties(sinon.spy(), {length: {value: 1}});

			cli.bind(CLI.head, spy);
			expect(() => cli.bind(CLI.head, spy)).to.throwError();
		});
	});

	describe("#bind(qwcli.lead, function)", () => {
		it("should return binding symbol", () => {
			const spy = sinon.spy();
			expect(typeof cli.bind(CLI.lead, spy)).to.be("symbol");
		});

		it("should bind positional arguments to handler", () => {
			const spy = defineProperties(sinon.spy(), {length: {value: 2}});

			cli.bind(CLI.lead, spy);
			const argv = cli.parser()(["baz", "bang", "biff"]);

			expect(spy.called).to.be(true);
			expect(spy.calledWith("baz", "bang")).to.be(true);
			expect(argv.length).to.be(1);
			expect(argv[0]).to.be("biff");
		});

		it("should bind positional arguments after options", () => {
			const leadspy = defineProperties(sinon.spy(), {length: {value: 1}});
			const foospy = sinon.spy();
			const barspy = sinon.spy();

			cli.bind(CLI.lead, leadspy);
			cli.bind("foo", foospy);
			cli.bind("bar", barspy);
			const argv = cli.parser()(["foo", "bar", "baz"]);

			expect(leadspy.called).to.be(true);
			expect(leadspy.calledWith("baz")).to.be(true);
			expect(foospy.called).to.be(true);
			expect(barspy.called).to.be(true);
			expect(argv.length).to.be(0);
		});

		it("should error on attempt to set lead twice", () => {
			const spy = defineProperties(sinon.spy(), {length: {value: 1}});

			cli.bind(CLI.lead, spy);
			expect(() => cli.bind(CLI.lead, spy)).to.throwError();
		});
	});

	describe("#bind(qwcli.rest, handler)", () => {
		it("should return binding symbol", () => {
			const spy = sinon.spy();
			expect(typeof cli.bind(CLI.rest, spy)).to.be("symbol");
		});

		it("should bind positional arguments to handler", () => {
			const spy = sinon.spy(argv => {
				expect(argv).to.be.an("array");
				expect(argv.length).to.be(2);
				expect(argv[0]).to.be("baz");
				expect(argv[1]).to.be("bang");
			});

			cli.bind(CLI.rest, spy);
			const argv = cli.parser()(["baz", "bang"]);

			expect(spy.called).to.be(true);
		});

		it("should bind positional arguments after lead", () => {
			const leadspy = defineProperties(sinon.spy(), {length: {value: 1}});
			const restspy = sinon.spy(argv => {
				expect(argv.length).to.be(1);
				expect(argv[0]).to.be("bar");
			});

			cli.bind(CLI.lead, leadspy);
			cli.bind(CLI.rest, restspy);
			const argv = cli.parser()(["foo", "bar"]);

			expect(leadspy.called).to.be(true);
			expect(leadspy.calledWith("foo")).to.be(true);
			expect(restspy.called).to.be(true);
		});

		it("should error on attempt to set rest twice", () => {
			const spy = sinon.spy();

			cli.bind(CLI.rest, spy);
			expect(() => cli.bind(CLI.rest, spy)).to.throwError();
		});
	});

	describe("#bind(RegExp, handler)", () => {
		it("should return binding symbol", () => {
			const spy = sinon.spy();
			expect(typeof cli.bind(/./, spy)).to.be("symbol");
		});

		it("should bind matching option to handler", () => {
			const spy = sinon.spy();

			cli.bind(/^f/, spy);
			const argv = cli.parser()(["foo", "fum", "bar"]);

			expect(spy.calledTwice).to.be(true);
			expect(argv.length).to.be(1);
			expect(argv[0]).to.be("bar");
		});

		it("should pass RegExp captures to handler", () => {
			const spy = defineProperties(sinon.spy(), {length: {value: 2}});

			cli.bind(/^(.*):(.*)$/, spy);
			const argv = cli.parser()(["foo:bar"]);

			expect(spy.called).to.be(true);
			expect(spy.calledWith("foo", "bar")).to.be(true);
			expect(spy.firstCall.args.length).to.be(3);
			expect(spy.firstCall.args[2]).to.be.a(CLI.State);
		});

		it("should pass qwcli.State to handler", () => {
			const spy = defineProperties(sinon.spy(), {length: {value: 1}});

			cli.bind(/^(.*):(.*)$/, spy);
			const argv = cli.parser()(["foo:bar"]);

			expect(spy.called).to.be(true);
			expect(spy.calledWith("foo")).to.be(true);
			expect(spy.firstCall.args.length).to.be(2);
			expect(spy.firstCall.args[1]).to.be.a(CLI.State);
		});

		it("should bind with options between head and lead", () => {
			const headspy = defineProperties(sinon.spy(), {length: {value: 1}});
			const leadspy = defineProperties(sinon.spy(), {length: {value: 1}});
			const foospy = sinon.spy();
			const respy = sinon.spy();

			cli.bind(CLI.head, headspy);
			cli.bind(CLI.lead, leadspy);
			cli.bind("foo", foospy);
			cli.bind(/^b/, respy);
			const argv = cli.parser()(["head", "foo", "bar", "foo", "lead"]);

			expect(foospy.calledTwice).to.be(true);
			expect(headspy.calledWith("head")).to.be(true);
			expect(leadspy.calledWith("lead")).to.be(true);
			expect(respy.called).to.be(true);
			expect(argv.length).to.be(0);
		});
	});

	describe("#bound(string)", () => {
		var foo, re;

		beforeEach(() => {
			foo = cli.bind("foo", sinon.spy());
			re = cli.bind(/^b/, sinon.spy());
		});

		it("should return the effective binding symbol for an option", () => {
			expect(cli.bound("foo")).to.be(foo);
			expect(cli.bound("bar")).to.be(re);
		});
	});

	describe("#handler(Binding)", () => {
		var binding, spy;

		beforeEach(() => {
			spy = sinon.spy();
			binding = cli.bind("foo", spy);
		});

		it("should return the handler for a binding", () => {
			expect(cli.handler(binding)).to.be(spy);
		});
	});

	describe("#plugin(Plugin)", () => {
		var cli, plugin;

		beforeEach(() => {
			cli = new CLI.Interface();
			plugin = sinon.spy();
		});

		it("should pass Interface to plugin", () => {
			cli.plugin(plugin);
			expect(plugin.calledWith(cli)).to.be(true);
		});

		it("should be chainable", () => {
			expect(cli.plugin(plugin)).to.be(cli);
		});
	});
});

describe("qwcli.stateful", () => {
	var cli, spy;

	beforeEach(() => {
		cli = new CLI.Interface();
		spy = defineProperties(sinon.spy(), {length: {value: 2}});
		spy[CLI.stateful] = true;
		cli.bind("foo", spy);
		cli.parser()(["foo", "bar"]);
	});

	it("should mark handler to receive final State argument", () => {
		expect(spy.calledOnce).to.be(true);
		expect(spy.firstCall.args.length).to.be(2);
		expect(spy.firstCall.args[0]).to.be("bar");
		expect(spy.firstCall.args[1]).to.be.a(CLI.State);
		expect(spy.firstCall.args[1].curr).to.be("foo");
	});
});

describe("qwcli.term", () => {
	var cli, foospy, barspy, leadspy;

	beforeEach(() => {
		foospy = sinon.spy();
		barspy = sinon.spy();
		leadspy = defineProperties(sinon.spy(), {length: {value: 1}});
		termspy = sinon.spy(() => CLI.term);

		cli = new CLI.Interface();
		cli.bind("foo", foospy);
		cli.bind("bar", barspy);
		cli.bind("--", termspy);
		cli.bind(CLI.lead, leadspy);
	});

	it("should terminate option parsing", () => {
		cli.parser()(["foo", "bar", "--", "foo"]);
		expect(foospy.calledOnce).to.be(true);
		expect(barspy.calledOnce).to.be(true);
		expect(termspy.calledOnce).to.be(true);
		expect(leadspy.calledOnce).to.be(true);
		expect(leadspy.firstCall.args.length).to.be(1);
		expect(leadspy.calledWith("foo")).to.be(true);
	});
});